/*---------------------------------------------------------------------------
 *  Copyright (C) 2007 Dallas Semiconductor Corporation, All Rights Reserved.
 * 
 *  Permission is hereby granted, free of charge, to any person obtaining a
 *  copy of this software and associated documentation files (the "Software"),
 *  to deal in the Software without restriction, including without limitation
 *  the rights to use, copy, modify, merge, publish, distribute, sublicense,
 *  and/or sell copies of the Software, and to permit persons to whom the
 *  Software is furnished to do so, subject to the following conditions:
 * 
 *  The above copyright notice and this permission notice shall be included
 *  in all copies or substantial portions of the Software.
 * 
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 *  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 *  IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 *  OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 *  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 *  OTHER DEALINGS IN THE SOFTWARE.
 * 
 *  Except as contained in this notice, the name of Dallas Semiconductor
 *  shall not be used except as stated in the Dallas Semiconductor
 *  Branding Policy.
 * ---------------------------------------------------------------------------
 */
/** 
 * \file lib8007.c
 * \brief Interface library for 8007 smartcard interface chips
 *
 * This library contains routines required to communicate with smartcards
 * via an 8007 while conforming to:
 * ISO 7816
 * EMV 4.1
 */

#include <stdio.h>
#include <string.h>

#include <absacc.h>
#include "common.h"
#include "lib8007.h"
#include "intrins.h"

#ifdef DS5250
    #include "reg5240.h"
#endif

#ifdef DS89C390
    #include "reg390.h"
#endif

#ifdef DS89C400
    #include "reg400.h"
#endif

#ifdef DS5002
	#include "reg5000.h"
#endif

#define  COLD_RESET 1
#define  WARM_RESET 2


#define  SPECIFIC_MODE_USE_DEFAULT_F_D   1
#define  SPECIFIC_MODE_USE_TA1_F_D       2
#define  NEGOTIABLE_MODE 				 3

#define T1_BLOCK_RETRY	3			/* Number of retries */

#define NONIBLOCK	0x80		/* Non I Block indicator */
#define SBLOCK		0x40		/* S Block indicator	 */
#define MOREBIT 	0x20		/* More bit indicator	 */

#define SBLOCKFUNCMASK	0x3F	/* S-block function mask */
	
#define CODESBLOCK(x)	    (unsigned char)((x)|NONIBLOCK|SBLOCK)
#define CODERBLOCK(nr,rc)   (unsigned char)(NONIBLOCK|((nr) << 4)|(rc))
#define CODEIBLOCK(ns,mb)   (unsigned char)(((ns) << 6)|((mb) << 5))

#define ISIBLOCK(pcb)	(!((pcb) & NONIBLOCK))			/* I block   */
#define ISRBLOCK(pcb)	((pcb & NONIBLOCK) && (!(pcb & SBLOCK))&&(!(pcb&MOREBIT)))	/* R block   */
#define ISSBLOCK(pcb)	(((pcb) & NONIBLOCK) && ((pcb) & SBLOCK))	/* S block   */
#define NR(pcb)		(((pcb) & 0x10) >> 4)	    /* R_Block Sequence number  */
#define NS(pcb)		(((pcb) & 0x40) >> 6)	    /* I_Block Sequence number  */
#define RERR(pcb) 	((pcb) & RERR_MASK)	    /* Error field in R-blck */
#define MORE(pcb) 	(((pcb) & MOREBIT) >> 5)	    /* More data bit */
#define SBLOCKFUNC(pcb)	((pcb) & SBLOCKFUNCMASK)	    /* S-block function */

#define SREQUEST_ABORT    0xC2 
#define SREQUEST_IFS      0xC1
#define SREQUEST_WTX      0xC3 

#define SRESPONSE_IFS     0xE1
#define SRESPONSE_WTX     0xE3 
#define SRESPONSE_ABORT   0xE2 

#define generateEDC(onebyte,value) (value^onebyte)

//returns active card slot working buffer starting address
#define ASSIGN_CURRENT_CARDSLOT_BUFFER()((currentSlot==0)?card1_workingBuffer:card2_workingBuffer);

int16_t  dssc_ATRsequence(uint8_t mode, uint8_t reset_type);

uint8_t  card1_workingBuffer[256];	   //Working buffer for card slot 1
uint8_t  card2_workingBuffer[256];	   //Working buffer for card slot 2

uint8_t  ATRLength[CARD_SLOTS];

uint8_t  PDR_value_for_Fi1_Di[4]={12,12,6,3};

uint8_t  GET_RESPONSE[5]={0,0xC0,0,0,0};

uint8_t  TMode[CARD_SLOTS];   // T=0 or T=1
uint8_t  ATRmode[CARD_SLOTS]; // specifies whether Card is in SPECIFIC_MODE or NEGOTIABLE_MODE
uint32_t WWT[CARD_SLOTS];     // Work wait time
uint32_t CWT[CARD_SLOTS];     // Character wait time
uint32_t BWT[CARD_SLOTS];     // Block wait time
uint8_t  EDCtype[CARD_SLOTS];
uint8_t  NAD[CARD_SLOTS];     // Node address byte to use when communicating in T=1
uint8_t  IFSC[CARD_SLOTS];    // Maximum segment length when sending to card in T=1

uint8_t  currentSlot = 0;	  // Contains active Slot number

uint8_t  ds8007_int=0;		  //This software interrupt flag gets set when interrupt handler
							  //gets called

uint8_t IFSD_Negotiated[CARD_SLOTS]={0,0}; //Flag indicates whether IFSD negotiation happened
uint8_t	IFD_Seq[CARD_SLOTS];  // Contains next Terminal I-block sequence number
uint8_t	ICC_Seq[CARD_SLOTS];  // Contains next Card I-block sequence number

uint8_t wtx[CARD_SLOTS]={0,0},t1_parity_error=0,t1_sblock_sent=0;	

#define reset_8007_interrupt_flag() ds8007_int=0; 

extern void writeBlock(uint8_t NAD, uint8_t PCB, uint8_t length, uint8_t *buffer,uint8_t EDC);

//returns version string
char *dssc_version(void)
{
	return (char *)DSSC_VERSION;
}

//ds8007 interrupt handler
void ds8007_intisr(void) interrupt 2
{
	ds8007_int=1;
	return;
} 

//reads byte from ds8007 uart in t1 mode
int16_t readT1Byte()
{
  uint8_t val;
  
  //wait for ds8007 interrupt
  while(!ds8007_int){ if((dssc_readregister(MSR)&MSR_TBE_RBF_MASK)==MSR_TBE_RBF_MASK) break;}

  reset_8007_interrupt_flag();
 
  val = dssc_readregister(USR);

  if ( val& USR_PE_MASK)
  {
    t1_parity_error=1;
  }

  if (val & (USR_TOL3_MASK|USR_TOL2_MASK|USR_TOL1_MASK))
  {
    return ERR_RECEIVE_TIMEOUT;
  }

  if(val&USR_TBE_RBF_MASK)
  {
  	// Read and store byte
  	return dssc_readregister(URR);
  }
  else
  {
   	return ERR_UNKNOWN_ERROR;
  }
}

//reads byte from ds8007 uart in t0 mode
int16_t readByte()
{
  uint8_t val;

  //wait for ds8007 interrupt
  while(!ds8007_int);
  reset_8007_interrupt_flag();
 
  val = dssc_readregister(USR);

  if ( val& USR_PE_MASK)
  {
  	if(TMode[currentSlot]==0)
	{
		dssc_powerdown();
	}
    return ERR_RECEIVE_PARITY;
  }

  if (val & (USR_TOL3_MASK|USR_TOL2_MASK|USR_TOL1_MASK))
  {
  	if(TMode[currentSlot]==0)
	{
		dssc_powerdown();
	}
    return ERR_RECEIVE_TIMEOUT;
  }

  if(val&USR_TBE_RBF_MASK)
  {
  	// Read and store byte
  	return dssc_readregister(URR);
  }
  else
  {
   	return ERR_UNKNOWN_ERROR;
  }
}


/*
	Write a byte and leave the 8007 in transmit mode
*/
int16_t writeByte(uint8_t onebyte)
{
  uint8_t val;

  val = dssc_readregister(UCR1);

  // set T bit
  dssc_writeregister(UCR1, (val | UCR1_T_R_MASK));

  reset_8007_interrupt_flag();

  dssc_writeregister(UTR, onebyte);

  // wait for ds8007 interrupt to happen
  while(!ds8007_int);

  reset_8007_interrupt_flag();

  val=dssc_readregister(USR);	

  if((val&USR_PE_MASK)==USR_PE_MASK)
  {
	dssc_powerdown();
	return ERR_RECEIVE_PARITY;
  }

  if(val& USR_TBE_RBF_MASK)  
  {
  	return 0;
  }
  else
  	return ERR_UNKNOWN_ERROR;
}

//Write a byte and configure the 8007 in receive mode after transmitting byte
int16_t writeLastByte(uint8_t onebyte)
{
  uint8_t val;

  //start 24 bit timer
  dssc_writeregister(TOC,0x7C);

  val = dssc_readregister(UCR1);
  // set LCT and T bit
  dssc_writeregister(UCR1,val | (UCR1_T_R_MASK));

  reset_8007_interrupt_flag();

  dssc_writeregister(UTR,onebyte);

  //wait for ds8007 interrupt
  while(!ds8007_int);
  reset_8007_interrupt_flag();

  val=dssc_readregister(USR);	

  //change to read mode
  dssc_writeregister(UCR1,(dssc_readregister(UCR1) &(~(UCR1_T_R_MASK))));

  if((val&USR_PE_MASK)==USR_PE_MASK)
  {
	dssc_powerdown();
	return ERR_RECEIVE_PARITY;
  }

  if(val & USR_TBE_RBF_MASK)
  { 
  	return 0;
  }
  else
  {
  	return ERR_UNKNOWN_ERROR;
  }
}
   
/*
Send a T=1 formatted block.
Formats T=1 packet from raw input data.  Length must already be
within the requirements of the destination smart card.
*/
void sendBlock(uint8_t NAD, uint8_t PCB, uint8_t length, uint8_t *buffer)
{
  uint8_t i;
  uint8_t EDC;
  uint32_t timeout;
  uint8_t t1,t2,t3,val;

  if(wtx[currentSlot]>0)
  {
  	timeout = (1L* BWT[currentSlot]*wtx[currentSlot])+1;  // Use BWT for T=1
  }
  else
  {
  	timeout = BWT[currentSlot]+1;  // Use BWT for T=1
  }

  t1=timeout;t2=timeout>>8;t3=timeout>>16;

  EDC = 0;

  EDC = generateEDC(NAD,EDC);
  EDC = generateEDC(PCB,EDC);
  EDC = generateEDC(length,EDC);

  // Set up timer for 24 bit timer, edge trigger start
  dssc_writeregister(TOC,0x00);
  dssc_writeregister(TOR3,t3);	
  dssc_writeregister(TOR2,t2);
  dssc_writeregister(TOR1,t1);

  timeout = CWT[currentSlot];  // Use CWT for T=1
  t1=timeout;t2=timeout>>8;t3=timeout>>16;

  for (i=0;i<length;i++)
  {
    EDC = generateEDC(buffer[i],EDC);
  }

  val=dssc_readregister(UCR1);

  // set T bit
  dssc_writeregister(UCR1, (val | UCR1_T_R_MASK));

  dssc_writeregister(TOC,0x7C);

  //wait for BGT bit to get set
  while((dssc_readregister(MSR)&MSR_BGT_MASK)!=MSR_BGT_MASK);

  writeBlock(NAD,PCB,length,buffer,EDC);
  reset_8007_interrupt_flag();	
  dssc_readregister(USR);

  //change to read mode
  dssc_writeregister(UCR1,(dssc_readregister(UCR1) &(~(UCR1_T_R_MASK))));
 
  dssc_writeregister(TOR3,t3);
  dssc_writeregister(TOR2,t2);
  dssc_writeregister(TOR1,t1);

  return;
}

/*
Receive a T=1 formatted block.
This does a raw receive, it does not check sequence numbers.
FIXME: The receive block function should do more to handle checking sequencing and
only return the raw data and some sort of pass/fail code to the caller.
*/
int16_t receiveBlock(uint8_t *rNAD, uint8_t *rPCB, uint8_t *rLEN)
{
  int16_t retval;
  int16_t index = 0;
  uint8_t EDC;
  int16_t i;
  uint8_t expectedLength;
  uint8_t *buffer;


  buffer=ASSIGN_CURRENT_CARDSLOT_BUFFER();

  t1_parity_error=0;

  // Get NAD, PCB, and Length
  retval = readT1Byte();
  if (retval < 0) { goto exit_from_receiveblock;}
  *rNAD = retval;
  retval = readT1Byte();
  if (retval < 0)  { goto exit_from_receiveblock;}
  *rPCB = retval;
  retval = readT1Byte();
  if (retval < 0)   { goto exit_from_receiveblock;}
  *rLEN = retval;

  // Add one to length for LRC
  expectedLength = *rLEN;

  // Add additional byte if using CRC

  // Get all data bytes plus EDC (1 byte at end)
  for (i = 0;i <= expectedLength;i++)
  {
    retval = readT1Byte();
    if (retval < 0)
    {
      	  goto exit_from_receiveblock;
    }
    buffer[index++] = retval;
  }

  EDC = 0;
  EDC = generateEDC(*rNAD,EDC);
  EDC = generateEDC(*rPCB,EDC);
  EDC = generateEDC(*rLEN,EDC);
  for (i = 0;i <= expectedLength;i++)
  {
      EDC = generateEDC(buffer[i],EDC);
  }

  if(t1_parity_error)
  {
   retval=ERR_RECEIVE_PARITY;
   goto exit_from_receiveblock;
  }

  if (EDC != 0)
  {
      retval=ERR_RECEIVE_LRC;
	  goto exit_from_receiveblock;
  }

  if (*rNAD != 0)
  {
      retval=ERR_INVALID_NAD;
	  goto exit_from_receiveblock;
  }

  if((ISRBLOCK(*rPCB)&&(*rLEN!=0))||(ISSBLOCK(*rPCB)&&(*rLEN>1)))
  {
  	retval=ERR_INVALID_LEN;
  }
  else
  {
  	retval=*rLEN;
  }

exit_from_receiveblock:


	EDC=dssc_readregister(UCR1);
	//read URR to clear FIFO. we dont want to retain the data in FIFO while going to transmit mode
	if((dssc_readregister(MSR)&MSR_TBE_RBF_MASK)==MSR_TBE_RBF_MASK) dssc_readregister(URR);
    dssc_writeregister(UCR1,(EDC |((UCR1_T_R_MASK))));
	dssc_writeregister(TOC,0x00);
	dssc_readregister(USR);

    if(*rLEN==0xFF)
      return ERR_ABORT_T1_TRANSACTION;

	return retval;
}

/*
Send an S block for setting IFSD (card terminal buffer size)
*/
int16_t dssc_sendsblockIFSD(uint8_t IFSD)
{
  uint8_t buffer[1];
  uint8_t rNAD,rPCB,rLEN;
  uint8_t *rbuffer;
  uint8_t count=T1_BLOCK_RETRY;  

  buffer[0] = IFSD;
  rbuffer=ASSIGN_CURRENT_CARDSLOT_BUFFER();

  do
  {
	  sendBlock(0,0xC1,1,buffer);
  
      if((receiveBlock(&rNAD,&rPCB,&rLEN))<0)
	  {
	    
	  	continue;
	  }

  	  // If we do not get a confirmation response, fail.
  	  if ((rPCB == 0xE1) && (rLEN == 1) && (rbuffer[0] == IFSD))
	  {
    	return 0;
	  }

   }while(--count);

   return ERR_SETIFSD_FAILURE;
}

/*
Set Node address info for this card
*/
void dssc_setNAD(uint8_t value)
{
  NAD[currentSlot] = value;
}

/*
	receives t1 protocol S block from card
*/
int receiveSBlock(uint8_t *rNAD, uint8_t *rPCB, uint8_t *rLEN)
{
  int ret; 
  int Responsed;
  uint8_t *rbuffer=ASSIGN_CURRENT_CARDSLOT_BUFFER();

  t1_sblock_sent=0;

  do
  {
    Responsed = 0;

    if((ret = receiveBlock(rNAD,rPCB,rLEN))<0)
	{
	 return ret;
	}

    switch (*rPCB)
	{
      case SREQUEST_ABORT :
	 // 	sendBlock(0,SRESPONSE_ABORT,0,(void *)0);
		return ERR_ABORT_T1_TRANSACTION;

      case SREQUEST_IFS :
	    if((*rbuffer<0x10)||(*rbuffer==0xff))
		{
		   return ERR_INVALID_INF;
		}

		sendBlock(0,SRESPONSE_IFS,1,rbuffer);
		IFSC[currentSlot] = *rbuffer;
		Responsed = 1;
		t1_sblock_sent=1;
        break;  

      case SREQUEST_WTX :
		 wtx[currentSlot]=*rbuffer;	 
         sendBlock(0,SRESPONSE_WTX,1,rbuffer);
		 wtx[currentSlot]=0;
         Responsed = 1;
		 t1_sblock_sent=1;
         break;
	}

  } while (Responsed);  
  return ret;
}


/*
Send a message using T=1 protocol
*/
int16_t dssc_sendAPDUT1(uint8_t *req_buf, uint16_t req_len, uint8_t *resp_buf)
{
  int ret;
  int currentIBlockLen,more,retry,i_block_retry;
  int totalResLength;
  uint8_t rNAD,rblock_sent=0;
  int cardiblock_err=0;
  uint8_t rPCB;
  uint8_t rLEN;
  uint8_t *IFDSeq, *ICCSeq;

  uint8_t *rbuffer=ASSIGN_CURRENT_CARDSLOT_BUFFER();

  IFDSeq=(IFD_Seq+currentSlot);
  ICCSeq=(ICC_Seq+currentSlot);

  if (IFSD_Negotiated[currentSlot]==0)
  {
    if((ret=dssc_sendsblockIFSD(0xFE))<0)
	{
		return ret;
	}
	IFSD_Negotiated[currentSlot] = 1;
  }

  while (req_len) 
  {
	currentIBlockLen = (req_len<IFSC[currentSlot])?req_len:IFSC[currentSlot]; 
  
    // We do chaining if request length is greater than IFSC
    req_len -= currentIBlockLen;
    more = req_len ? 1 : 0;     
    
    retry = T1_BLOCK_RETRY;

    i_block_retry=1;
	cardiblock_err=0;

	sendBlock(0,CODEIBLOCK(*IFDSeq,more), currentIBlockLen,req_buf);
	
	//make sure block gets sent successfully within three times
    while (retry--)
	{

      ret = receiveSBlock(&rNAD,&rPCB,&rLEN);


      if ((ret >= 0 ))
	  {

          if((ISRBLOCK(rPCB)))
		  {

		  		if((NR(rPCB) == *IFDSeq)&&(!t1_sblock_sent))
				{  
				    if(((++i_block_retry)>3))
					{
						retry=-1;
						ret=ERR_MAX_RETRIED;
						break;
					}

			 		sendBlock(0,CODEIBLOCK(*IFDSeq,more), currentIBlockLen, req_buf);
					retry++;  
	        		continue;
		 		}
				else 
				{	if (more)
					{
            			*IFDSeq = 1 - *IFDSeq; 
		        		break;  
					}
					else if(((!rblock_sent)&&(retry<2))) //there could be error in icc response to last i-block
					{
						sendBlock(0,CODERBLOCK(*ICCSeq, ((cardiblock_err==ERR_RECEIVE_LRC)||(cardiblock_err==ERR_RECEIVE_PARITY)) ? 1 : 2), 0,(void*)0);
						retry++;
						rblock_sent=1;
						continue;
					}
		 		}

	   		}
			else if (ISIBLOCK(rPCB))
			{ 
    	  		if (((NS(rPCB) == *ICCSeq) && (more==0)))
				{
					*IFDSeq = 1 - *IFDSeq;
					break; 
		  		}
	   		}
	  	} 

	   if(ret== ERR_ABORT_T1_TRANSACTION)
	   {
			return ERR_ABORT_T1_TRANSACTION;
 	   }

		if(retry)
		{
			if((cardiblock_err)) ret=cardiblock_err;
	  		//send R block saying we got invalid block
	  		sendBlock(0,CODERBLOCK(*ICCSeq, ((ret==ERR_RECEIVE_LRC)||(ret==ERR_RECEIVE_PARITY)) ? 1 : 2), 0,(void*)0);
			
			cardiblock_err =ret;
		}
	
	}



    if (retry < 0)
	{
		return ERR_MAX_RETRIED;    
	}
    req_buf += currentIBlockLen;
  }

  totalResLength = 0;


  do
  {
    memcpy(resp_buf,rbuffer, rLEN);
    resp_buf += rLEN;
    totalResLength += rLEN;
	
    *ICCSeq = 1 - *ICCSeq;

    if (MORE(rPCB))
	{
	  sendBlock(0,CODERBLOCK(*ICCSeq, 0), 0, (void*)0);
	  retry = T1_BLOCK_RETRY;

      while (retry)
	  {

        ret = receiveSBlock(&rNAD,&rPCB,&rLEN);

	    if(ret >= 0)
		{

          if (ISRBLOCK(rPCB))
		  {
	         if (!(NR(rPCB) != *IFDSeq))
			 {
			 	/* resent our R-block */
                sendBlock(0,CODERBLOCK(*ICCSeq, 0), 0,(void*)0);
				continue;  
			 }
		  }
		  else if (ISIBLOCK(rPCB))
		  {
			  if (NS(rPCB) == *ICCSeq)
			  {
				     break;  /* A valid I-Block,break this loop */
			  }	
		  }
		}

		if(ret== ERR_ABORT_T1_TRANSACTION)
		{
			return ERR_ABORT_T1_TRANSACTION;
		}

		if(--retry)
		{
	  		//send R block saying we got invalid block
			sendBlock(0,CODERBLOCK(*ICCSeq, 0), 0,(void*)0);
		}

	  }

	  if (!retry)
	  {
	   	return ERR_MAX_RETRIED;
	  }
	}
	else
	{
		break;
	}
  } while (1);
	return totalResLength;
}

/*
Send a message using T=0 protocol
*/
int16_t dssc_sendAPDUT0(uint8_t *buffer, uint16_t length, uint8_t *rbuffer)
{
  int index;
  uint16_t rindex = 0, Lc,Le;
  uint8_t val;
  uint8_t INS,P3=0,wait_for_proc_byte,*hdr,SW1, SW2, warning_status_byte=0,case4_cmd=0,get_resp=0;
  int retval;

  uint8_t wlt1,wlt2,wlt3, rt1,rt2,rt3;
  uint32_t timeout;
  timeout=WWT[currentSlot];


  rt3=(timeout>>16);
  rt2=(timeout>>8);
  rt1=(timeout);

  timeout++;

  wlt3=(timeout>>16);
  wlt2=(timeout>>8);
  wlt1=(timeout);

  Lc=Le=0;

  dssc_writeregister(TOC,0x00);
  
  if(length>5)
  {
  	Lc=buffer[4];
	if(length>Lc+5)
	{     //check whether it is case3 command or case4 command
  		Le=buffer[length-1];
		case4_cmd=1;
	}
    P3=Lc;
  }
  else if(length==5)	//check whether it is case2 command or case1 command
  {
  	Le=buffer[4];
	P3=Le;
  }

  hdr=buffer;

  reset_8007_interrupt_flag();
  dssc_readregister(USR);

  //state machine to process t=0 apdu request and response
  while(1)
  {	

    INS=hdr[1];
	wait_for_proc_byte=1;

  	// Write 5 byte command header
  	for (index = 0;index < 4;)
  	{
    	retval=writeByte(hdr[index++]);
	    if (retval < 0)
    	{
      		return retval;
    	}
  	}

   //set work wait timeout after last byte
   dssc_writeregister(TOR3,wlt3);
   dssc_writeregister(TOR2,wlt2);
   dssc_writeregister(TOR1,wlt1);

	retval=writeLastByte(P3);

	if (retval < 0)
    {
     	return retval;
    }
	
	index=5;

  //set work wait timeout after reading one byte
  dssc_writeregister(TOR3,rt3);
  dssc_writeregister(TOR2,rt2);
  dssc_writeregister(TOR1,rt1);


  	while (wait_for_proc_byte)
    {
	
    	// Get procedure byte
    	retval = readByte();

	    if (retval < 0)
    	{
      		return retval;
    	}

		val=retval;

		switch(val)
		{
			case 0x61:
					  get_resp=1;
			case 0x6C:
					  Lc=0;
					  hdr=(val==0x61)?GET_RESPONSE:buffer;
					  P3=Le=retval=readByte();
					  Le+=rindex;
					  if(retval<0) return;
					  wait_for_proc_byte=0;
					  //wait at least four etus. 
					  //character wait time between opposite direction should be atleast 16 etus.
					  while((dssc_readregister(MSR)&MSR_BGT_MASK)!=MSR_BGT_MASK);

			case 0x60: 
					  break;
			default:
				if((val==INS)||(val==~INS))
			 	{
			 	  if((Lc+5)==index)
				  {
        			  while((Le>rindex) && ((retval = readByte()) >= 0))
        			  {
          			    rbuffer[rindex++] = retval;
				 	    if(val==~INS)
						   break;
				      }
				      if (retval < 0)
						return retval;
				 	  break;					  
				    }

				 	//wait at least four etus.
				 	//character wait time between opposite direction should be atleast 16 etus.
				    while((dssc_readregister(MSR)&MSR_BGT_MASK)!=MSR_BGT_MASK);

				  	if(val==INS)
				  	{
			      		for (;index < (Lc+4);index++)
				  		{
	          			  retval = writeByte(buffer[index]);
				      
						  if (retval < 0)
    					  {
      						return retval;
    					  }
				  		}
				  	}

				  //set work wait timeout after writing last byte
				  dssc_writeregister(TOR3,wlt3);
				  dssc_writeregister(TOR2,wlt2);
				  dssc_writeregister(TOR1,wlt1);
       				  
  			  	  retval=writeLastByte(buffer[index++]);

				  //set work wait timeout after reading one byte
				  dssc_writeregister(TOR3,rt3);
				  dssc_writeregister(TOR2,rt2);
				  dssc_writeregister(TOR1,rt1);


				      
				  	if (retval < 0)
    			  	{
      					return retval;
    			  	}
				  	break;
			  	}

			  if(((val&0xf0)!=0x60)&&((val&0xf0)!=0x90))
			  {
			  	dssc_powerdown();
				return ERR_INVALID_STATUS_BYTE;
			  }

			  retval=readByte();
			  if (retval < 0) return retval;

			  if(!warning_status_byte)
			  {
				  SW2=retval;	
 			      SW1=val;
			  	  if((((SW1==0x62)||(SW1==0x63))||(((SW1&0x90)==0x90)&&(!((SW1==0x90)&&(SW2==0)))))&&((Lc+5)==index)&&(case4_cmd)&&(get_resp==0))
				  {
					  warning_status_byte=1;

					  Lc=0;
					  hdr=buffer=GET_RESPONSE;
					  P3=Le=0;
					  Le+=rindex;
					  wait_for_proc_byte=0;

					  //character wait time between opposite direction should be atleast 16 etus.
					  while((dssc_readregister(MSR)&MSR_BGT_MASK)!=MSR_BGT_MASK);

				  }
				  else
				  {
					  rbuffer[rindex++]=SW1;
   					  rbuffer[rindex++]=SW2;
					  return rindex;
				  }	 
			  }
			  else
			  {
					  if(rindex!=0)
					  {
					  	rbuffer[rindex++]=SW1;
   					  	rbuffer[rindex++]=SW2;
					  }
					  else
					  {
  					  	rbuffer[rindex++]=val;
   					  	rbuffer[rindex++]=retval;
					  }
					  return rindex;
			  }

      	}
     }
  }
}

/*
Send a message using T=0 or T=1 protocol
*/
int16_t dssc_sendAPDU(uint8_t *buffer,uint16_t length,uint8_t *rbuffer)
{
  int ret;
  switch (TMode[currentSlot])
  {
    case 0:
      return dssc_sendAPDUT0(buffer,length,rbuffer);
      break;
    case 1:
      if((ret=dssc_sendAPDUT1(buffer,length,rbuffer))<0)
	  {
	   dssc_powerdown();
	   return ret;
	  }
      break;
    default:
      return ERR_PROTOCOL_UNSUPPORTED;
      break;
  }
}

/*
	performs card warm reset and receives ATR from Card
*/
int16_t dssc_warmreset(uint8_t mode)
{
  uint8_t val;

  // Check for power status
  val = dssc_readregister(PCR);
  if (!(val & PCR_START_MASK))
    return ERR_POWERUP_VOLTAGE_INVALID;

  // Apply reset
  val = dssc_readregister(PCR);
  dssc_writeregister(PCR,val & ~PCR_RSTIN_MASK);

  // Call common getATR routine
  return dssc_ATRsequence(mode, WARM_RESET);
}

/*
	performs card powerup and receives ATR from Card
*/
int16_t dssc_powerup(uint8_t mode, uint8_t voltage)
{
  uint8_t val;

  // Compile time setting of operating crystal frequency
#if ((CRYSTAL_FREQUENCY_8007 >= 1000000L) && (CRYSTAL_FREQUENCY_8007 < 3000000L))
  // Set smartcard clock to 1/1 crystal
  dssc_writeregister(CCR,0x00);
#elif ((CRYSTAL_FREQUENCY_8007 >= 3000000L) && (CRYSTAL_FREQUENCY_8007 < 6000000L))
  // Set smartcard clock to 1/2 crystal
  dssc_writeregister(CCR,0x01);
#elif ((CRYSTAL_FREQUENCY_8007 >= 6000000L) && (CRYSTAL_FREQUENCY_8007 < 10000000L))
  // Set smartcard clock to 1/4 crystal
  dssc_writeregister(CCR,0x02);
#elif ((CRYSTAL_FREQUENCY_8007 >= 10000000L) && (CRYSTAL_FREQUENCY_8007 <= 40000000L))
  // Set smartcard clock to 1/8 crystal
  dssc_writeregister(CCR,0x03);
#else 
  // Set smartcard clock to 1/2 internal oscillator (about 1.44MHz)
  // NOTE: Can only change CCR.2 when shifting to internal oscillator
  dssc_writeregister(CCR,0x01);
  dssc_writeregister(CCR,0x05);

  // Wait for internal oscillator to engage (check the MSR.CLKSW bit on page 18)
  do
  {
    val = dssc_readregister(MSR);
  }
  while (!(val & MSR_CLKSW_MASK));
#endif

  // Set the power supply voltage
  val = dssc_readregister(PCR);

  // Clear 1.8V and 3V bits
  dssc_writeregister(PCR, val & ~(PCR_3V_5V_MASK|PCR_1V8_MASK));

  switch(voltage)
  {
    case POWERUP_5V:
      // Do nothing
      break;
    case POWERUP_3V:
      val = dssc_readregister(PCR);
      // Set 3V bit
      dssc_writeregister(PCR,(val|PCR_3V_5V_MASK));
      break;
    case POWERUP_1p8V:
      val = dssc_readregister(PCR);
      // Set 1.8V bit
      dssc_writeregister(PCR,(val|PCR_1V8_MASK));
      break;
    default:
      return ERR_POWERUP_VOLTAGE_INVALID;
      break;
  }

  // Apply reset
  val = dssc_readregister(PCR);
  dssc_writeregister(PCR,val & ~PCR_RSTIN_MASK);

  val = dssc_readregister(HSR);
  do
  {
    // Power the card, RST low, C4 and C8 high
    val = dssc_readregister(PCR);
    dssc_writeregister(PCR,val | (PCR_C8_MASK|PCR_C4_MASK|PCR_START_MASK));
    val = dssc_readregister(HSR);
    if (val & (HSR_PRTL2_MASK|HSR_PRTL1_MASK|HSR_PRL2_MASK|HSR_PRL1_MASK|HSR_PTL_MASK))
    {
      dssc_powerdown();
      return ERR_POWERUP_INTERRUPTED;
    }
    val = dssc_readregister(PCR);

  }
  while (!(val & PCR_START_MASK));

  // Call common getATR routine
  return dssc_ATRsequence(mode, COLD_RESET);
}


/*
	receives ATR from Card
*/
int16_t dssc_ATRsequence(uint8_t mode, uint8_t reset_type)
{
  int  jcount;

  uint8_t val;
  uint8_t USRval;
  uint8_t index;
  uint8_t etucount = 0;
  int16_t i;
  uint8_t historicalBytes;
  uint8_t expectedCharacters, t0_only_atr;
  uint8_t interfaceIteration = 1;
  uint8_t done = 0;
  uint8_t check;
  uint8_t reject_atr=0;
  uint16_t TB3=-1, status=0;
  uint8_t TCK, TA1=0x0, TC1=0, TC2=0x0A, TB1=0,D=1;

  uint8_t *workingBuffer;

  // Default to T=0 mode
  uint8_t T = 0;
  uint8_t curByte;

  //intialize default protocol configuration
  NAD[currentSlot] = 0;
  IFSC[currentSlot] = 0x20;  
  CWT[currentSlot] = 11 + (1 << 13);
  BWT[currentSlot] = 11L + (1L << 4L) * 960L;
  ATRmode[currentSlot] = NEGOTIABLE_MODE;
  t0_only_atr=1;

  workingBuffer=ASSIGN_CURRENT_CARDSLOT_BUFFER();

  // Reset the UART
  val = dssc_readregister(CSR);
  dssc_writeregister(CSR, val & ~CSR_nRIU_MASK);

  // Remove UART reset
  val = dssc_readregister(CSR);
  dssc_writeregister(CSR, val | CSR_nRIU_MASK);

  // Set FIFO to 1 byte
  dssc_writeregister(FCR, 0x00);

  // Set divisor
  dssc_writeregister(PDR, 12);

  // Set prescaler and
  // Enable auto convention
  dssc_writeregister(UCR2, 0);

  // Set SS bit
  dssc_writeregister(UCR1, UCR1_SS_MASK);

  reset_8007_interrupt_flag();

  jcount = 0;

  // Wait 40000 to 45000 cycles to release reset
  dssc_writeregister(TOC,0x00);
  dssc_writeregister(TOR3,0x00);
  dssc_writeregister(TOR2,0x6C);
  dssc_writeregister(TOR1,0x00);
  dssc_writeregister(TOC,0x61);

  while(!ds8007_int);
  reset_8007_interrupt_flag();
  
  val=dssc_readregister(USR);

  if (mode == POWERUP_ISO)
  {
    // Wait up to 40000 cycles for ATR to start
    dssc_writeregister(TOC,0x00);
    dssc_writeregister(TOR3,0x00);
    dssc_writeregister(TOR2,0x7D);
    dssc_writeregister(TOR1,0xF0);
  }
  else  // power up EMV
  {
    // Wait up to 40000 cycles for ATR to start and 20160 etu counter after first byte
    dssc_writeregister(TOC,0x00);
    dssc_writeregister(TOR3,0x00);
    dssc_writeregister(TOR2,0x7D);
    dssc_writeregister(TOR1,0xFC);
  }

  // Release reset
  val = dssc_readregister(PCR);
  dssc_writeregister(PCR,val | PCR_RSTIN_MASK);
  dssc_writeregister(TOC,0x65);	  //start timer3


  //wait for ds8007 interrupt
  while(!ds8007_int);
  val = dssc_readregister(USR);
  if (val & (USR_PE_MASK|USR_FER_MASK|USR_OVR_MASK|USR_EA_MASK))
  {
        dssc_powerdown();
        return ERR_POWERUP_ATR_INVALID; 
  }
  
  if (val & USR_TOL3_MASK)
  {
      dssc_powerdown();
      return ERR_POWERUP_ATR_TIMEOUT;
  }

  if (mode == POWERUP_ISO)
  {
    // reset etucount to calculate for 9600 etu between characters in ATR
	etucount=0;
  }
  else  // mode is EMV
  {
    // Set up Timer2:Timer3 for maximum of 20160 etu for whole ATR
    dssc_writeregister(TOC,0x05);
    dssc_writeregister(TOR3,0x4E);
    dssc_writeregister(TOR2,0xB6);
    // Start timer
    dssc_writeregister(TOC,0x65);

	//reset etucount to calculate for 10080 etu between characters in ATR, and
	etucount=0;
  }

  index = 0;

  while (!done)
  {
	//wait for ds8007 interrupt to happen

	while(!ds8007_int);
	reset_8007_interrupt_flag();

    // Read UART status
	if(index) 
	{
    	USRval = dssc_readregister(USR);
	}else
	{
		USRval=val;
	}

    {
      // If receive buffer full
      if (USRval & USR_TBE_RBF_MASK)
      {
        // Read and store ATR byte
        val = dssc_readregister(URR);
        curByte = val;
        workingBuffer[index++] = val;
        if (index == 1)
        {
	      if ((curByte != 0x3f) && (curByte != 0x3b))
		  {
  			dssc_powerdown();
        	return ERR_POWERUP_ATR_INVALID;
		  }
		  check = 0;
        }
		else
		  check ^= curByte;

        if (index == 2)
        {
          historicalBytes = curByte & 0x0F;
          expectedCharacters = curByte & 0xF0;
        }
        if (index > 2)
        {
          switch(expectedCharacters)
          {
            case 0x00:
              // Historical characters
              historicalBytes--;
              if (historicalBytes == 0)
              {
                if(t0_only_atr)
                {
                  done = 1;
                }
                else
                  expectedCharacters = 0x01;  // Go to checksum state
              }
              break;
            case 0x01:
              // TCK case
              TCK = curByte;
              done = 1;
              break;

            case 0x10:
            case 0x30:
            case 0x50:
            case 0x70:
            case 0x90:
            case 0xB0:
            case 0xD0:
            case 0xF0:
              // TA case
              expectedCharacters &= 0xE0;
			  switch(interfaceIteration)
			  {
			  	//TA1 case
			  	case 1:
						if((curByte<0x11)||(curByte>0x13))
						{
							reject_atr=1;
						}
						TA1=curByte;
						D=TA1&0xf;
						D=(D==3)?4:D;
						break;
				case 2:
			  			if((curByte&0xf)!=T)
						{
							reject_atr=1;
						}
						else
						{
						 	if(curByte&0x10)
							{
								ATRmode[currentSlot]=SPECIFIC_MODE_USE_DEFAULT_F_D;
								
								if(mode==POWERUP_EMV)
								{
									reject_atr=1;
								}
							}
							else
								ATRmode[currentSlot]=SPECIFIC_MODE_USE_TA1_F_D;
						}
						break;
				case 3:
						if(mode==POWERUP_EMV)
						{
							if(((curByte>=0x00) && (curByte<0x10))||(curByte==0xff))
							{
						 	   reject_atr=1;
							}
						}
						IFSC[currentSlot]=curByte;
			  }
              break;

            case 0x20:
            case 0x60:
            case 0xA0:
            case 0xE0:
              // TB case
              expectedCharacters &= 0xD0;

			  if(mode==POWERUP_EMV)
			  {
			  	switch(interfaceIteration)
			  	{
			  		case 1:
						if((reset_type==COLD_RESET)&&(curByte!=0))
						{
							reject_atr=1;
						}
						TB1=1;
						break;
					case 2:
							reject_atr=1;
						break;
	 		    }
			  }
			  if(interfaceIteration==3)
			  {
				TB3=curByte;							  
			  }
              break;

            case 0x40:
            case 0xC0:
              // TC case
              expectedCharacters &= 0xB0;
			  switch(interfaceIteration)
			  {
			  	case 1: 
						TC1=curByte;
						if((reset_type==COLD_RESET)&&(!TB1))
						{
							reject_atr=1;
						}
						break;
				case 2:
						if((mode==POWERUP_EMV)&&(curByte!=0x0A))
						{
							reject_atr=1;
						} 
						else
						{
							TC2=curByte;
						}
						break;
				case 3:
						if(curByte!=0x00)
						{
							reject_atr=1;
						} 
			  }
              break;

            case 0x80:
              // TD case
              expectedCharacters=(curByte&0xF0);
              // Handle zero historical characters
              // If we get TD1, we have the first protocol selection
              if ((interfaceIteration==1))
              {
                T=curByte & 0x0F;
			  	if((curByte&0xF)>1)
			  	{
				   reject_atr=1;
			  	}

				if(((reset_type==COLD_RESET)&&(!TB1)))
				{
					reject_atr=1;
				}

              }
			  else if(interfaceIteration==2)
			  {
			  	if((((curByte&0x0F)|T)!=0xE)&&((curByte&0x0F)!=1))
				{
					reject_atr=1;
				}
			  }

			  if((curByte&0x0F)>0)
				   t0_only_atr=0;

              interfaceIteration++;
              break;

            default:
			  reject_atr=1;
            break;
          }

          if ((expectedCharacters == 0x00) && (historicalBytes == 0))
          {
                if (t0_only_atr)
                {
                  done = 1;
                }
                else
	              expectedCharacters = 0x01;
          }
        }

        // Reset etu counter to start calculating inter character timeout
 	    etucount=0;
      }
    }

	if(USRval&(USR_PE_MASK|USR_FER_MASK))
	{
      dssc_powerdown();
	  dssc_writeregister(TOC,0x00);
      return ERR_POWERUP_ATR_INVALID;
	}

    // Read Timer status
    if((mode == POWERUP_EMV) && (USRval & USR_TOL3_MASK))
    {
      dssc_powerdown();
	  dssc_writeregister(TOC,0x00);
	  return ERR_POWERUP_ATR_TIMEOUT;
    }

    if ((USRval & USR_TOL1_MASK))
    {
      // If 240 ETU rollover for ISO or 252 ETU rollover for EMV, increment total count
      etucount++;
      // If we exceed 240*40=9600 etus for ISO or 252*40=10080 etus for EMV, we fail as interchar timeout happened
      if (etucount == 40)
      {
       	dssc_powerdown();
 	    dssc_writeregister(TOC,0x00);
	    return ERR_POWERUP_ATR_TIMEOUT;
      }
    }
  }

  dssc_writeregister(TOC,00); //stop timer0	

  // If we are in T=1 protocol, we get a LRC (xor checksum) at the end
  if (T==1)
  {
    if ((check != 0)&&(reject_atr==0))
    {
	  dssc_powerdown();
	  return ERR_POWERUP_ATR_CRC_FAILURE;
    }
  }

  if(mode==POWERUP_EMV)
  {
  	if(T==1)
	{
		i=TC1;
		i=(TC1==0xff)?-1:TC1;
		if((TB3==-1)||((TB3&0x0F)>0x05)||((TB3&0xF0)>0x40)||((1 << (TB3&0x0f))<=(i+1)))
		{
			reject_atr=1;
		}
	}
  }

  if(reject_atr==1)
  {

	  if(reset_type==WARM_RESET)
	  {
       	dssc_powerdown();
      	return ERR_POWERUP_ATR_INVALID;;
	  }
	  else
	  {
		return ERR_DO_WARM_RESET;
	  }
  }
  
  // Set ATR length for the ATR received
  ATRLength[currentSlot] = index;
  // Set T protocol mode
  TMode[currentSlot] = T;

  // Set the 8007 UCR1.PROT mode according to T value
  val = dssc_readregister(UCR1);
  if (T == 0)
    dssc_writeregister(UCR1,val & ~UCR1_PROT_MASK);
  else
    dssc_writeregister(UCR1,val | UCR1_PROT_MASK);

  // Set extra guard time
  dssc_writeregister(GTR, TC1);

  //set new PDR value
  if(ATRmode[currentSlot]==SPECIFIC_MODE_USE_TA1_F_D)
  {
  	dssc_writeregister(PDR,PDR_value_for_Fi1_Di[(TA1&0x0F)]);
  }
  else
  {
      D=1;
  }

  // Set WWT 
  WWT[currentSlot] = TC2 * 960 *D;

  if(mode==POWERUP_EMV)
  {
  	WWT[currentSlot]+=480*D;
  }
 
  // Set CWT and BWT if present in TB3
  if (TB3 != -1)
  {
      val = TB3 & 0x0F;
      CWT[currentSlot] = 11 + (1 << val);
      val = (TB3 & 0xF0) >> 4;
      BWT[currentSlot] = 11L + (1L << val) * 960L * D;
  }

  if(T==1)
  {
  	if(mode==POWERUP_EMV)
	{
		BWT[currentSlot]+=960L*D;
		CWT[currentSlot]+=4;

		//Initialize T=1 protocol variables
		IFD_Seq [currentSlot] = ICC_Seq [currentSlot] = IFSD_Negotiated[currentSlot]= 0;
	}
  }
  else
  {
 	dssc_writeregister(FCR,0x40); 
  }
  return 0;
}

/*
	Powers down the Card
*/
int16_t dssc_powerdown()
{
  uint8_t val;

  // Power down the card
  val = dssc_readregister(PCR);
  dssc_writeregister(PCR,(val & ~PCR_START_MASK));
  return 0;
}


/*
	Selects given card interface in ds8007 chip
*/
int16_t dssc_selectcard(uint8_t slot)
{
  uint8_t val;

  val = dssc_readregister(CSR);
  switch (slot) 
  {
    case 1:
      // Select appropriate smartcard and clear other card selects
      dssc_writeregister(CSR,val & ~(CSR_SC3_MASK|CSR_SC2_MASK) | CSR_SC1_MASK);
      break;
    case 2:
      // Select appropriate smartcard and clear other card selects
      dssc_writeregister(CSR,val & ~(CSR_SC3_MASK|CSR_SC1_MASK) | CSR_SC2_MASK);
      break;
    case 3:
      // Select appropriate smartcard and clear other card selects
      dssc_writeregister(CSR,val & ~(CSR_SC2_MASK|CSR_SC1_MASK) | CSR_SC3_MASK);
      break;
    default:
      return ERR_INVALID_SLOT;
  }

  // Set global slot identifier value
  currentSlot = slot-1;

  return 0;
}

/*
	Checks card presence for given card slot
*/
int16_t dssc_checkpresence(uint8_t slot)
{
  uint8_t val;

  val = dssc_readregister(MSR);
  
  switch(slot)
  {
    case 1:
      if (val & MSR_PR1_MASK) return 1;
      break;
    case 2:
      if (val & MSR_PR2_MASK) return 1;
      break;
    default:
      return ERR_INVALID_SLOT;
      break;
  }
  return 0;
}



/*
	Checks card presence for given card slot
*/
uint8_t dssc_checkprotocol(slot)
{
   return TMode[slot];
}

/*
Library initialization
*/
int16_t dssc_init()
{
  uint8_t val;

  // Reset the UART
  dssc_writeregister(CSR, 0x00);

  // Select smart card slot 1
  dssc_writeregister(CSR, CSR_SC1_MASK);

  // Make sure card is powered down
  val = dssc_readregister(PCR);
  dssc_writeregister(PCR, val & ~PCR_START_MASK);

  // Select smart card slot 2
  dssc_writeregister(CSR, CSR_SC2_MASK);

  // Make sure card is powered down
  val = dssc_readregister(PCR);
  dssc_writeregister(PCR, val & ~PCR_START_MASK);

  // Deselect all cards
  dssc_writeregister(CSR, 0x00);

  //read USR and HSR register to clear the ds8007 interrupts
  dssc_readregister(USR);
  dssc_readregister(HSR);

  //read USR and HSR register to clear the ds8007 interrupts
  dssc_readregister(USR);
  dssc_readregister(HSR);

  return 0;
}

/*
	Copies current card slot ATR to given input buffer
*/
int16_t dssc_getATRbuffer(uint8_t *buff, int16_t length)
{
  uint8_t *workingBuffer;
  workingBuffer=ASSIGN_CURRENT_CARDSLOT_BUFFER();
  memcpy(buff, workingBuffer, (length<ATRLength[currentSlot])?length:ATRLength[currentSlot]);
  return ATRLength[currentSlot];
}


/*
	The following are accessary functions to read and write data byte to ds8007 registers
	this functions will be used in Microcontrollers other than ds5002. for ds5002, quite optimized
	assembly functions have been implemented to increase performance and meet EMV timings.
*/

#ifndef DS5002

uint8_t dssc_readregister(uint8_t address)
{
  idata uint8_t rAddress, toReturn ;
  bit ea_bit;
  rAddress = address;

#ifdef DS5250
	ea_bit=_testbit_(EA);
	DMOS |= 0x01;
#endif

#ifdef DS5002
	ea_bit=_testbit_(EA);
	RPCTL|=0x20; 				//open access across extended bus
#endif

  switch (rAddress)
  {
    case URR:
      // If reading URR, check the CRED bit in MSR so that
      // we don't overrun the 8007.  See MSR.CRED bit description on page
      // 18 of 8007 spec.
    while (!(BASE_ADDRESS[MSR] & MSR_CRED_MASK))
    {
        // If card de-powers exit with zero return value.
        if ((BASE_ADDRESS[PCR] & 0x01) == 0)
        {

#ifdef DS5250
    DMOS &= 0xFE;
    EA=	ea_bit;
#endif

#ifdef DS5002
	RPCTL &= 0xDF;
    EA=	ea_bit;
#endif
          return 0;
        }
      }
      break;
    default:
      break;
  }
  toReturn = BASE_ADDRESS[rAddress];

#ifdef DS5250
DMOS &= 0xFE;
EA=1;
#endif

#ifdef DS5002
	RPCTL &= 0xDF;				//close access across extended bus;
	EA=1;
#endif

  return toReturn;
}

void dssc_writeregister(uint8_t address, uint8_t value)
{
  idata uint8_t rAddress, rValue;
  bit ea_bit;

  rAddress = address;
  rValue = value;

#ifdef DS5250
	ea_bit=_testbit_(EA);
DMOS |= 0x01;
#endif

#ifdef DS5002
	ea_bit=_testbit_(EA);
	RPCTL|=0x20; 				//open access across extended bus
#endif

  switch (rAddress)
  {
    case TOC:
    case UTR:
      // If writing TOC or UTR, check the CRED bit in MSR so that
      // we don't overrun the 8007.  See MSR.CRED bit description on page
      // 18 of 8007 spec.
      while (!(BASE_ADDRESS[MSR] & MSR_CRED_MASK));
      break;
    default:
      break;
  }
  BASE_ADDRESS[rAddress] = rValue;

#ifdef DS5250
	DMOS &= 0xFE;
    EA=	ea_bit;
#endif

#ifdef DS5002
	RPCTL &= 0xDF;				//close access across extended bus;
    EA=	ea_bit;
#endif
}
#endif